package checkers.interned;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;

import checkers.flow.Flow;
import checkers.quals.Interned;
import checkers.types.*;
import static checkers.types.AnnotatedTypeMirror.*;

import checkers.util.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

/**
 * An {@link AnnotatedTypeFactory} that accounts for the properties of the
 * Interned type system. This type factory will add the {@link Interned}
 * annotation to a type if the input:
 * 
 * <ul>
 * <li>is a String literal
 * <li>is a class literal
 * <li>is an enum constant
 * <li>has a primitive type
 * <li>has the type java.lang.Class
 * <li>is a call to the method {@link String#intern()}
 * </ul>
 */
public class InternedAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Adds annotations from tree context before type resolution. */
    private final TreePreAnnotator treePre;

    /** Adds annotations from the resulting type after type resolution. */
    private final TypePostAnnotator typePost;

    /** The {@link Interned} annotation. */
    final AnnotationMirror INTERNED;

    /** Flow-sensitive qualifier inference. */
    private final Flow flow;

    /**
     * Creates a new {@link InternedAnnotatedTypeFactory} that operates on a
     * particular AST.
     * 
     * @param env the current processing environment for this annotation
     *        processor
     * @param root the AST on which this type factory operates
     * @param useFlow whether or not flow-sensitive qualifier inference should
     *        be used
     */
    public InternedAnnotatedTypeFactory(ProcessingEnvironment env,
        CompilationUnitTree root, boolean useFlow) {
        super(env, root);
        this.INTERNED = annotations.fromName("checkers.quals.Interned");
        this.treePre = new TreePreAnnotator();
        this.typePost = new TypePostAnnotator();
        
        this.flow = new Flow(env, root, INTERNED, this);
        if (useFlow) flow.scan(root, null);
    }
    
    @Override
    protected void annotateImplicit(Element elt, AnnotatedTypeMirror type) {
        typePost.visit(type);
    }

    @Override
    protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
        treePre.visit(tree, type);
        if (flow.test(tree) == Boolean.TRUE) 
            type.addAnnotation(INTERNED);
        typePost.visit(type);
    }
    
    /**
     * A class for adding annotations to a type based on its tree context.
     */
    protected class TreePreAnnotator extends
        SimpleTreeVisitor<Void, AnnotatedTypeMirror> {
        
        @Override
        public Void visitLiteral(LiteralTree node, AnnotatedTypeMirror type) {
            // Literals: add an @Interned annotation.
            type.addAnnotation(INTERNED);
            return null;
        }
    }

    /** 
     * A class for adding annotations to a type after initial type resolution.
     */
    protected class TypePostAnnotator extends AnnotatedTypeScanner<Void, Void> {

        @Override
        public Void visitDeclared(AnnotatedDeclaredType t, Void p) {
            
            // Enum types and constants: add an @Interned annotation.
            Element elt = types.asElement(t.getUnderlyingType());
            if (elt != null && elt.getKind() == ElementKind.ENUM)
                t.addAnnotation(INTERNED);

            // Annotate class types.
            TypeElement classElt = elements.getTypeElement("java.lang.Class");
            assert classElt != null;            
            if (types.isSameType(types.erasure(t.getUnderlyingType()),
                    types.erasure(classElt.asType())))
                t.addAnnotation(INTERNED);

            return super.visitDeclared(t, p);
        }
        
        @Override
        public Void visitPrimitive(AnnotatedPrimitiveType t, Void p) {
            
            // All primitive types are implicitly interned.
            t.addAnnotation(INTERNED);
            
            return super.visitPrimitive(t, p);
        }

        @Override
        public Void visitExecutable(AnnotatedExecutableType t, Void p) {

            // Annotate the java.lang.String.intern() method.
            if (t.getElement() != null) {
                ExecutableElement method = t.getElement();
                if (method.getSimpleName().contentEquals("intern")) {
                    Name clsName = ElementUtils.getQualifiedClassName(method);
                    if (clsName != null && clsName.contentEquals("java.lang.String"))
                        t.getReturnType().addAnnotation(INTERNED);
                }
            }

            // skip the receiver
            scan(t.getReturnType(), p);
            scanAndReduce(t.getParameterTypes(), p, null);
            scanAndReduce(t.getThrownTypes(), p, null);
            scanAndReduce(t.getTypeVariables(), p, null);
            return null;
        }
    }
    
    @Override
    public AnnotatedPrimitiveType getUnboxedType(AnnotatedDeclaredType type) {
        AnnotatedPrimitiveType primitive = super.getUnboxedType(type);
        primitive.addAnnotation(INTERNED);
        return primitive;
    }
}
