package checkers.nonnull;

import javax.lang.model.element.*;

import checkers.basetype.*;
import checkers.source.Result;
import checkers.types.AnnotatedTypeMirror;

import com.sun.source.tree.*;

/**
 * A type-checking visitor for the {@link checkers.quals.NonNull} type
 * qualifier that uses the {@link BaseTypeVisitor} implementation. This visitor
 * reports errors or warnings for violations for the following cases:
 *
 * <ul>
 * <li>if the receiver of a member dereference is not NonNull (error
 * "dereference.of.nullable")</li>
 * </ul>
 */
public class NonNullVisitor extends BaseTypeVisitor<Void, Void> {

    private final AnnotationMirror NONNULL, RAW;

    /**
     * Creates a new visitor for type-checking {@link checkers.quals.NonNull}.
     *
     * @param checker the checker to use
     * @param root the root of the input program's AST to check
     */
    public NonNullVisitor(NonNullChecker checker, CompilationUnitTree root) {
        super(checker, root);
        NONNULL = this.annoFactory.fromName("checkers.quals.NonNull");
        RAW = this.annoFactory.fromName("checkers.quals.Raw");
    }

    @Override
    public Void visitMemberSelect(MemberSelectTree node, Void p) {
       
        AnnotatedTypeMirror type = factory.getAnnotatedType(node.getExpression());
        
        final Element select = factory.elementFromUse(node);
        assert select != null;
        
        if (select.getKind() != ElementKind.METHOD
                && (type.hasAnnotation(RAW) || !type.hasAnnotation(NONNULL)))
            checker.report(Result.failure("dereference.of.nullable",
                        node.getExpression().toString()), node);

        return super.visitMemberSelect(node, p);
    }

}
